import debug from 'debug';
import { EventEmitter } from 'events';
import IORedis, { Redis } from 'ioredis';
import { from, fromEvent, Observable } from 'rxjs';
import { filter, map, switchMap, takeWhile, tap } from 'rxjs/operators';
import { ApplicationException } from '../exceptions/application.exception';
import { PubSubOptions } from './interfaces/pub-sub-options.interface';
import {
  PubSubRawMessage,
  PubSubRawNextMessage,
} from './interfaces/pub-sub-raw-message.interface';

const log = debug('fennec:pubsub:instance');
export class PubSub extends EventEmitter {
  pubRedis: Redis;
  pSubRedis: Redis;
  channels: string[] = [];
  event$: Observable<[string, string, any]>;

  constructor(private readonly options: PubSubOptions) {
    super();
    this.pubRedis = new IORedis(this.options.redis);
    this.pSubRedis = new IORedis(this.options.redis);

    this.event$ = fromEvent<[string, string, string]>(
      this.pSubRedis,
      'pmessage',
    ).pipe(
      map((ev) => {
        try {
          ev[2] = JSON.parse(ev[2]);
        } catch (err) {
          log('WARN: is not json');
          return null;
        }
        log('on message: %s %s %o', ...ev);
        return ev;
      }),
      filter((v) => !!v),
    );
  }

  async disconnect() {
    log('disconnecting to redis...');
    this.pubRedis.disconnect();
    this.pSubRedis.disconnect();
    log('disconnected');
  }

  async registerChannel(channel: string) {
    if (this.channels.includes(channel)) {
      return;
    }
    this.channels.push(channel);

    return await this.pSubRedis.psubscribe(channel);
  }

  private async redisPublish<T>(channel: string, message: PubSubRawMessage<T>) {
    log.extend('publish')('channel: %s, message: %O', channel, message);
    return await this.pubRedis.publish(channel, JSON.stringify(message));
  }

  async publish(channel: string, message: any): Promise<number> {
    return await this.redisPublish(channel, {
      type: 'next',
      message,
    });
  }

  async finish(channel: string): Promise<number> {
    return await this.redisPublish(channel, {
      type: 'complete',
    });
  }

  async throwError(channel: string, error: string): Promise<number> {
    return await this.redisPublish(channel, {
      type: 'error',
      error,
    });
  }

  message$ = <T>(channel: string): Observable<T> => {
    return from(this.registerChannel(channel))
      .pipe(switchMap(() => this.event$))
      .pipe(
        filter(([pattern]) => pattern === channel),
        tap(([pattern, channel, message]) => {
          log.extend('subscribe')(
            'channel: %s, match: %, message: %O',
            channel,
            pattern,
            message,
          );
        }),
        takeWhile(([pattern, channel, message]) => {
          if (pattern === channel) {
            if (message.type === 'error') {
              throw new ApplicationException(message.error);
            }
            return message.type !== 'complete';
          }
          return true;
        }),
        map((ev) => ev[2] as PubSubRawMessage<T>),
        map((message: PubSubRawNextMessage<T>) => message.message),
      );
  };
}
